#!/usr/bin/perl

#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#***********************************************************************
#	This script currently supports the following videosites:
#		[*] youtube - de, com, fr
#		[*] myvideo
#		[*] clipfish
#		[*] metacafe
#		[*] vimeo
#		[*] video.google
#***********************************************************************
#		If you receive an error like
#		"user@host ~ $ Can't locate XYZ.pm in @INC (@INC contains:)..."
#		you'll have to install the appropriate perl modules
#		On debian-based systems, you'll need to install 
#		the following packages:
#			libwww-perl
#			libgetopt-long-descriptive-perl
#			libmp3-info-perl
#			libterm-progressbar-perl
#
#		Alternative:  use "perl -MCPAN -e shell"...
#***********************************************************************
#		vim-settings:
#			set tabstop=4
#			set shiftwidth=4
#***********************************************************************

use strict;
use warnings;
use Getopt::Long;
use LWP::UserAgent;
use MP3::Info;
use Term::ProgressBar;
use FileHandle;
#use Data::Dumper;

Getopt::Long::Configure ("bundling", "gnu_compat");
STDOUT->autoflush;
$SIG{INT} = \&catcher;		#Catch Ctrl-C and exit

my %INFO = (
	authors => "Haui",
	contact => "haui45\@web.de",
	date	=> "2009-01-02",
	version => "v0.9.0a",
	license => "GPL",
	name	=> "yaydl - yet another youtube downloader",
);

#*********************************************************************
my $videopath = "/my/videopath/";	#IMPORTANT - change to your path!*
#*********************************************************************

my @site_content;	#contains the html sourcecode
my $path = "";		#
my $encode = "";	#output file for the encoding proces
my $source = "";	#input file for the encoding process
my $name = "";		#new filename
my $ua;				#LWP user agent
my $filehandle ;	#filehandle for syswrite

my $currentsize = 0;# stuff
my $bit = 0;		# for
my $percent = 0;	# the
my $progress;		# progressbar

my $lasturl = "";	#used to discover redirections on video.google.com
my $sigint = 0;

my %options = (
	rename	=> '0',
	help	=> '0',
	sound	=> '0',
	version	=> '0',
	encode	=> '0',
	print	=> '0',
	tag		=> '0',
	quiet	=> '0',
	stdin	=> '0',
);

sub catcher() {
	$sigint = 1;
	close $filehandle;
	die "SIGINT recieved....exiting\n";
}
sub checkargs(){
	#parse the commadline options
	GetOptions ("rename|r" => \$options{'rename'}, 
				"help|h" => \$options{'help'},
				"sound|s" => \$options{'sound'}, 
				"version|v" => \$options{'version'}, 
				"encode|e" => \$options{'encode'},
			   	"print" => \$options{'print'},
				"tag|t"=>\$options{'tag'},
				"path|p=s" => \$videopath,
				"stdin" => \$options{'stdin'},
				"quiet|q"=>\$options{'quiet'}
				);

	if ($options{'version'} == 1){
		version();
		$options{'version'} = 0;
		exit 0;
	}
	if ($options{'help'} == 1){
		help();
		$options{'help'} = 0;
		exit 0;
	}

	if ($#ARGV +1 == 0){
		unless ($options{'stdin'} == 1){
			print "No URL specified\n";
			exit;
		}
		readurls();
		$options{'quiet'} = 1;
	}
	if ($videopath =~ m#/my\/videopath/# or $videopath eq ""){
		$videopath = $ENV{"HOME"} || 
			die "Please set a proper videopath!\n";
	}
	$videopath .= "/" unless ($videopath =~ m#/$#);
	unless (-e $videopath){
		warn "Videopath doesn't exist!\n";
		exit(1);
	}
#print "@{[ %options ]}\n";
}

sub readurls {
	foreach my $arg (<STDIN>){
		chomp $arg;
		$arg =~ s#\s+##g; #remove trailing/leading whitespaces
		next if ($arg =~ m#^$#);
		push @ARGV, $arg;
	}
}

#Replacement for the `which`-systemcalls..
#we don't need an extra perl module for this ;-)
#returns 0 on success
#		 1 else
sub which {
	my $programname = shift;
	my $path = $ENV{'PATH'} or return 1;
	my @paths = split(/:/, $path);
	foreach $path (@paths){
		$path .= "/$programname";
		if (-e $path){
			return 0;
		}
	}
	return 1;
}



# calls the methods for the given urls
sub choose(){
	my ($path, $url,$dec);
	unless ($options{'quiet'}){
		print "\nImportant notice: All links must be quoted correctly ->" . 
		" \"http://www.youtube.com...\"\n";
		print "Type \"y\" to proceed\n";
		print "[y/n]\n";
		$dec = <STDIN>;
		chomp $dec;
		unless ($dec =~ m/^y(es)?$/i){
			warn "Quitting...\n";
			exit 1;
		}
	}
	$ua = LWP::UserAgent->new();
	$ua->agent('Mozilla/5.0');
	foreach my $link (@ARGV){
		last if ($sigint == 1);
		print "---\n";
		if ($link =~ m#^(http://)?(www\.|de\.|fr\.)?youtube\.com/#){
			($url, $path) = youtube($link);
		}
		elsif ($link =~ m#^http://(www\.)?myvideo\.de\/watch#){
			($url, $path) = myvideo($link);
		}
		elsif (($link =~ m#^(http://)?(www\.)?clipfish\.de/#)){
			($url, $path) = clipfish($link);
		}
		elsif (($link =~ m#^(http://)?(www\.)?metacafe.com/watch/.*#)){
			($url, $path) = metacafe($link);
		}
		elsif (($link =~ m#^http://beta.vreel.net/watch_.*#)){
			($url, $path) = vreel($link);
		}
		elsif (($link =~ m#^(http://)?(www.)?vimeo.com/#)){
			($url, $path) = vimeo($link);
		}
		elsif (($link =~ m#^(http://)?(www\.)?video\.google\.(de|com)/.*docid=.*#)){
			($url, $path) = google($link);
		}
		else{
			warn "$link: Not a valid Link!\n";
			next;
		}
		if ($url eq ""){
			next;
		}
		#just print the downloadlink...
		if($options{'print'}){
			print "Downloadlink: $url\n";
			next;
		}

		warn "Download in progress - this may take a while\n";
		if (download($url, $path) != 0){
			die "Error: Couldn't get $url\n\n";
		}
		warn "\n[***************]Saved file $path [***************]\n\n";

		if($options{'encode'} == 1){
			if (encode("$path") == -1){

			}
			else{
				warn "\n[**************] Encoded file $encode [****************]\n\n" unless 
				($options{'rename'} == 1);
			}
		}
		if($options{'sound'} == 1){
			if (sound("$path") == -1){
				die "ripping error!\n";
			}
			warn "\n[**************] Encoded file $encode [****************]\n\n" unless 
			($options{'rename'} == 1);
		}
		if ($options{'sound'} == 1 || $options{'encode'} == 1 ){
			unlink("$path");
		}
		else{
			if ($options{'rename'} == 1){
				print "\nEnter a new filename\n";
				$name = <STDIN>;
				chomp($name);
				$name = "$videopath" . "/$name";
				rename("$path", "$name") or
				die "renaming failed!\n\n";
				warn "\n$path renamed successfully to $name\n\n";
			}
			$path="";
		}

	}
	print "---\n";
}

################
#   youtube    #
################
sub youtube() {
	my $url = shift;
	$url = "http://" . $url unless ($url =~ m#^http://#);
	#the base-url for the download
	my $line = "http://youtube.com/get_video?video_id="; 
	if ((getpage($url)) != 0){
		warn "Error: Couldn't get $url\n\n";
		return "";
	}
	my @testme = grep (/^.*var swfArgs.*, "t": ".*",.*/, @site_content);
	unless(defined $testme[0]){
		warn "Error: testme[0] undefined!\n\n";
		return "";
	}
	my $vidid = $testme[0];
	my $id = $testme[0];
	$id =~ s/^.*, "t": "//;		#delete everything before the long videoid
	$id =~ s/",.*//;			#delete everything after the long videoid 
	$vidid =~ s/.*"video_id": "//;	# get the short videoid
	$vidid =~ s/",.*//;				#
	$path = "$videopath" . "/$vidid" . ".flv"; # build the downloadpath
	$url = $line . $vidid . "\&t=" . $id;	# concat the download-url 
	return $url, $path;
}

################
#   google    #
################
sub google() {
	my $url = shift;
	my $copy = $url;
	
	$url = "http://" . $url unless ($url =~ m#^http://#);
	if ($url =~ m#http://video.google.com/videosearch.*#){
		warn "This link seems to be a preview page. Please  " .
			"select the right video and rerun yaydl!\n";
		return "";
	}
	
	if ((getpage($url)) != 0){
		warn "Error: Couldn't get $url\n\n";
		return "";
	}
	$url =~ s#.*docid=([-\d\w]*)&.*#$1#;
	$path = $videopath . $url . ".flv";
	
	#test, if the video is embedded from an other site
	# => extract the original source an append it to our videolist -
	# maybe it's supported by this script :)
	my @external_source = grep(/document.getElementById\('external_page'\).src = '/, @site_content);
	if (defined $external_source[0]){
		$external_source[0] =~ s#.*\('external_page'\)\.src = '(.*)';#$1#;
		#replace all occurrences of hexadecimal numbers like "\x3d"
		#with their corresponding ascii-symbol
		$external_source[0] =~ s#(\\x..)#chr(hex substr($1, 2, 2))#eg;
		print "enqueued $external_source[0]\n";
		push @ARGV, $external_source[0];
		return "";
	}
	#redirection handling
	if ($copy ne $lasturl){
		push @ARGV, $lasturl;
		print "enqueued $lasturl\n";
		return "";
	}

	my @testme = grep (/<a href="http:\/\/vp.video.google.com\/videodownload.*secureurl=.*>/, @site_content);

	#the site doesn't contain a downloadlink - 
	#maybe it's a link from the search-and-preview-site
#	unless(defined $testme[0]){
#		undef $url;
#		#check if we can find the link to the right video
#		@testme =  grep(/<div id="res-0" class="rl-res"/, @site_content);
#		$url = $testme[0];
#		#no? -> fail!
#		unless (defined $url){
#			warn "Error: testme[0] undefined!\n\n";
#			return "";
#		}
#		$url =~ s#.*docid="([^"]*)".*#$1#;
#		$url = "http://video.google.com/videoplay?docid=" . $url;
#		print "enqueued $url\n";
#		push @ARGV, $url;
#		return "";
#	}
	my $vidid = $testme[0];
	$vidid =~ s#.*<a href="http://vp.video.google.com/videodownload#http://vp.video.google.com/videodownload#;
	$vidid =~ s#">.*##s;
	$url = $vidid ;
	return $url, $path;
}


################
#    vimeo     #
################
sub vimeo(){
	my $url = shift;
	$url = "http://" . $url unless ($url =~ m#^http://#);
	#extract the first video-id and
	#download the page, that contains
	#the timestamp and the request_signature for the 
	#video
	$url =~ s#(http://)?(www\.)?vimeo.com/(\d*).*#$3#;
	my $vidid = $url;
	$path = $videopath . $vidid . ".flv";
	$url = "http://vimeo.com/moogaloop/load/clip:" . $vidid;
	print "$url\n";
	if ((getpage($url)) != 0){
		warn "Error: Couldn't get $url\n\n";
		return "";
	}
	my $request_sig = (grep(/request_signature/, @site_content))[0];
	my $expires = (grep(/request_signature_expires/, @site_content))[0];
	$request_sig =~ s#.*<request_signature>(.*)</request_signature>.*#$1#s;
	$expires =~ s#.*<request_signature_expires>(.*)</request_signature_expires>.*#$1#s;

	#that's is the downloadlink for the video :-)
	$url = "http://vimeo.com//moogaloop/play/clip:" . $vidid 
		. "/" . $request_sig . "/" . $expires . "/?q=sd";

	return $url, $path;
}

################
#    vreel     #
################
sub vreel() {
	my $url = shift;
	$url = "http://" . $url unless ($url =~ m#^http://#);
	if ((getpage($url)) != 0){
		warn "Error: Couldn't get $url\n\n";
		return "";
	}
	$url =~ s#(.*watch_)(.*)(\.html)#$2#; # get the videonumber
	$path = $videopath . $url . ".flv";

	# extract the downloadlink from the http-page
	my @testme = grep (/^.*embed type=.*llnwd.*\.avi.*/, @site_content);
	unless(defined $testme[0]){
		warn "Error: Couldn't get $url\n\n";
		return "";
	}
	my $id = $testme[0];
	$id =~ s#.*src="##;	# get the downloadlink for the video
	$id =~ s#" .*##;	#
	chomp $id;
	$url = $id;
	return $url, $path;

}

################
#   clipfish   #
################
sub clipfish() {
	my ($url, $serverid) ="";
	$url = shift;
	$url = "http://" . $url unless ($url =~ m#^http://#);
	if ((getpage($url)) != 0){
		print "Error while getting given url!\n\n";
		return "";
	}
	#get the first videoid 
	my @array = grep /var video\s*=\s/, @site_content;
	unless (defined $array[0]){
		print "array[0] undefined...maybe the site doesn't contain a video?\n\n";
		return "";
	}
	#extract the first videoid
	$array[0] =~ s#^.*id:\s*(\d*) , .*#$1#;
	my $vidid = $array[0];

	my $server = "http://www.clipfish.de/video_n.php?p=0|DE&vid=" . 
					$vidid;

	#we retrieve a second webpage, that contains
	#the downloadlink for the video
	if ((getpage($server)) != 0){
		print "Error: failed to get the encrypted id!\n\n";
		return "";
	}
	my $id = $site_content[0];
	$id =~ s#^.*http#http#;
	$id =~ s#flv.*#flv#;
	$url = $id;
	$path = $videopath . $vidid . ".flv";

	return $url, $path;
}

################
#   metacafe   #
################
sub metacafe(){
	my $url = shift;
	$url = "http://" . $url unless ($url =~ m#^http://#);
	my $path = $url;
	if ((getpage($url)) != 0){
		print "Error: Couldn't get $url\n\n";
		return "";
	}
	my @ids = grep (/so\.addParam\("flashvars",/, @site_content);
	unless (defined $ids[0]){
		print "Error: grep failed!\n\n";
		return "";
	}
	#pretty ugly....
	$url = $ids[0];
	$url =~ s#.*mediaURL=##;
	$url =~ s#&postRollContentURL.*##;
	$url =~ s#(%..)#chr(hex substr($1, 1, 2))#eg;
	$url =~ s#&#\?#g;
	$url =~ s#gdaKey#__gda__#;
	$path =~ s#(http://)?(www\.)?metacafe.com/watch/##;
	$path =~ s#/.*##;
	$path = $videopath . $path . ".flv";
	return $url, $path;
}

################
#   myvideo    #
################
sub myvideo(){
	my ($path, $url);
	my @swfargs;
	my $vidid = shift;
	$vidid = "http://" . $vidid unless ($vidid =~ m#^http://#);

	if ((getpage($vidid)) != 0){
		print "Error: Couldn't get $url\n\n";
		return "";
	}

	$vidid =~ s#(http://)?(www\.)?myvideo\.de/watch/##;
	$vidid =~ s#/.*##;
	$path = "$videopath" . "/$vidid" . ".flv";

	@swfargs = grep /SWFObject\(/, @site_content;
	my @server = grep /<link rel='image_src' href='.*\.jpg.*><link/ , @site_content;
	unless (defined($swfargs[0]) && defined($server[0])) {
		print "variable swfargs/server undefined...stopping method\n\n"; 
		return "";
	}
	my $id=$swfargs[0];
	#take the link of an image and extract the server-location - 
	#not very nice, but it works ;)
	$server[0] =~ s#.*link rel='image_src' href='##;
	$server[0] =~ s#thumbs/.*.jpg.*##;

	$id =~ s#.*p.addVariable\('_videoid','##;
	$id =~ s#'\).*##;
	chomp($id);
	$url = $server[0] . $id . ".flv";
	return $url, $path;

}

sub download() {
	my ($url, $path) = @_;
	print "Downloalink: $url\n";
	print "File: $path\n";
	my ($request, $response);
	my $chunksize=1024;
	my $filesize=0;
	$request = HTTP::Request->new('HEAD', $url);			#	Filesize
	$response = $ua->request($request);
	#HEAD response ok!
	if ($response->is_success()){
		if (int($response->header('Content-Length')) == 0){
			print "Filesize 0 byte....stopping method!\n\n";
			return -1;
		}
		print "Filesize: " . int($response->header('Content-Length')/1024)."kb\n";
		$filesize=$response->header('Content-Length');
	}
	#HEAD failed
	else{
		#try simple download
		warn "!!!\nFallback to simple download - the progressbar doesn't work in this mode!\n!!!\n";
		$request = HTTP::Request->new('GET', $url);
		$response = $ua->request($request, $path);
		if ($response->is_success()){
			return 0;
		}

		print "Error code   : ", $response->code() ,    "\n";
		print "Error message:  ", $response->message() , "\n";
		return -1;
	}
	$bit = int ($response->header('Content-Length')/100); 
	$request = HTTP::Request->new('GET', $url);
	open $filehandle, '>', $path or die $!;
	$progress = Term::ProgressBar->new (100);
	$response = $ua->request($request, \&callback, $chunksize);

	if ( $response->is_error() ) {
		print "Error code    : ", $response->code() ,    "\n";
		print "Error message:  ", $response->message() , "\n";
		return 1;
	}
	close $filehandle;	
	$percent = 0;
	$currentsize = 0;
	$bit = 0;
	undef $progress;
	undef $filehandle;
	print "\n";

	return 0;
}

sub callback() {
	my $data = shift;
	my $length = length $data;
	$currentsize += $length;
	if($currentsize >= $percent*$bit){
		$percent++;
		$progress->update ($percent-1);
	}
	syswrite $filehandle, $data, $length; 
}

sub getpage() {
	my $link = shift;

	my ($path, $request, $response, $url) = "";
	$request = HTTP::Request->new('GET', $link);
	$response = $ua->request($request);
	print "Retrieving http-page...\n";
	if ( $response->is_error() ) { 
		print "Error code    : ", $response->code() ,    "\n";
		print "Error message:  ", $response->message() , "\n";
		return -1;
	}
	else {
		#print Dumper($response);
		@site_content = split /\n/,  $response->content();
		$lasturl = $response->request->uri;
		return 0, ;
	}
}

sub encode(){
	$source = shift;
	$encode = $source;
	$encode =~ s/\.flv/\.avi/;
	$encode =~ s/\.divx/\.avi/;
	my $ffmpeg = which "ffmpeg";
	my $mencoder = which "mencoder";
	my @commands;
	my $return;
	if (! $mencoder){
		push(@commands, "mencoder \"$source\" -oac mp3lame -lameopts abr:br=92" .
			" -ovc xvid -xvidencopts bitrate=150 -o \"$encode\"");
		print "mencoder: found\n";
	}
	if(! $ffmpeg){
		push (@commands, "ffmpeg -i \"$source\" \"$encode\"");
		print "ffmpeg: found\n";
	}
	if($ffmpeg && $mencoder){
		$options{'encode'} = 0;
		print "mencoder: not found!\nffmpeg: not found!\nencoding disabled\n";
		return -1;
	}
	foreach my $command(@commands){
		$return = system($command);
		if ($return == 0){
			print "success\n";
			last;
		}
	}
	if ($return != 0){
		print "error - encoding disabled\n";
		$options{'encode'} = 0;
		return -1;
	}

	if ($options{'rename'} == 1){
		print "\nEnter a new filename\n";
		$name = <STDIN>;
		chomp($name);
		$name =~ s/.avi$//;
		$name = "$videopath" . "/$name" . ".avi";
		rename("$encode", "$name") or 
				die "Error: encode rename! :(\n";
		print "\n$encode renamed successfully to $name\n\n";
	}
}

sub sound(){
	$source = shift;
	$encode = $source;
	$encode =~ s/\.flv/\.mp3/;
	$encode =~ s/\.divx$/\.mp3/;
	system("mplayer -dumpaudio -dumpfile \"$encode\" \"$source\"") == 0
		or return -1;
	if ($options{'rename'} == 1){
		print "\nEnter a new filename\n";
		$name = <STDIN>;
		chomp($name);
		$name =~ s/.mp3$//;
		$name = "$videopath" . "/$name" . ".mp3";
		rename("$encode", "$name") or 
				die "Error: sound rename! :(\n";
		print "\n$encode renamed successfully to $name\n\n";
	}
	if ($options{'tag'} == 1){
		eval {
			print "Enter new title: ";
			my $title = <STDIN>;
			print "Enter new artist: ";
			my $artist = <STDIN>;
			print "Enter new album: ";
			my $album = <STDIN>;
			chomp($title, $artist, $album);
			my $mp3 = new MP3::Info "$name";
			$mp3->title($title);
			$mp3->artist($artist);
			$mp3->album($album);
			print "tagging successful\n\n";
		};
		if($@){
			print "tagging error\n";
			return -1;
		}
	}
	return 0;
}


sub help(){
	print "Usage: yaydl  [-esrtphv] \"link1\" \"link2\" ...\n".

	"\nImportant notice: all links must be quoted correctly!\n".
	"Parameters:\n".
	"   -e, --encode \t Encode the videos from *.flv to *.avi - ". 
			"requires mencoder or ffmpeg \n".
	"    \t\t\t yaydl -e \"http://youtube...\"\n".
	"   -s, --sound\t\t Extract the soundtracks of the downloaded videos - ".
			"requires mplayer!\n".
	"    \t\t\t yaydl -s \"http://youtube...\"\n".
	"   -r, --rename \t Rename the downloaded files\n" .
	"   -t, --tag\t\t Provides a basic way for tagging the extracted soundfiles\n" .
	"   -p, --path PATH \t Change the default downloadpath to PATH\n" . 
	"       --print \t\t Display the downloadlinks and exit\n" . 
	"   -q, --quiet\t\t Don't wait for the user's confirmation\n".
	"       --stdin\t\t Read URLs from STDIN (activates --quiet)\n".
	"   -h, --help   \t Show this help.\n".
	"   -v, --version \t Print version.\n";

	return;
}
sub version(){
	print "yaydl - yet another youtube downloader \nversion $INFO{'version'}".
   		"	- $INFO{'date'}\n";
	print "Author: $INFO{'authors'}, contact $INFO{'contact'}\n";
	return;
}

checkargs();
choose();

